"""In-process libmspack bindings for CAB, CHM, and SZDD files."""

from __future__ import annotations

from . import _libmspack as _libmspack  # noqa: F401
from .cab import CabArchive, CabFileInfo, CabInfo
from .chm import ChmArchive, ChmFileInfo, ChmInfo
from .kwaj import KwajFile, KwajInfo
from .errors import (
    CabDecompressionError,
    CabError,
    CabFormatError,
    CabPathTraversalError,
    ChmDecompressionError,
    ChmError,
    ChmFormatError,
    ChmPathTraversalError,
    KwajDecompressionError,
    KwajError,
    KwajFormatError,
    KwajPathTraversalError,
    MspackDecompressionError,
    MspackError,
    MspackFormatError,
    MspackPathTraversalError,
    SzddDecompressionError,
    SzddError,
    SzddFormatError,
    SzddPathTraversalError,
)
from .szdd import SzddFile, SzddInfo

__all__ = [
    "CabArchive",
    "CabFileInfo",
    "CabInfo",
    "ChmArchive",
    "ChmFileInfo",
    "ChmInfo",
    "SzddFile",
    "SzddInfo",
    "KwajFile",
    "KwajInfo",
    "MspackError",
    "MspackFormatError",
    "MspackDecompressionError",
    "MspackPathTraversalError",
    "CabError",
    "CabFormatError",
    "CabDecompressionError",
    "CabPathTraversalError",
    "ChmError",
    "ChmFormatError",
    "ChmDecompressionError",
    "ChmPathTraversalError",
    "SzddError",
    "SzddFormatError",
    "SzddDecompressionError",
    "SzddPathTraversalError",
    "KwajError",
    "KwajFormatError",
    "KwajDecompressionError",
    "KwajPathTraversalError",
]

__version__ = "0.2.0"



from pathlib import Path

import pytest

from pylibmspack import KwajFile, KwajFormatError, KwajPathTraversalError


FIXTURES = Path(__file__).resolve().parent / "fixtures"


def test_kwaj_info_and_read(tmp_path):
    kwj = KwajFile(str(FIXTURES / "sample.kwj"))
    info = kwj.info()
    assert "compression" in info
    assert info["data_offset"] >= 0
    assert info["headers"] >= 0
    assert info["extra_length"] >= 0

    data = kwj.read()
    out_path = kwj.extract(str(tmp_path))
    assert Path(out_path).read_bytes() == data


def test_kwaj_from_bytes():
    data = (FIXTURES / "sample.kwj").read_bytes()
    kwj = KwajFile.from_bytes(data, name="sample.kwj")
    assert kwj.read()


def test_kwaj_invalid_bytes():
    kwj = KwajFile.from_bytes(b"not a kwaj", name="bad.kwj")
    with pytest.raises(KwajFormatError):
        kwj.info()


@pytest.mark.parametrize(
    "name",
    [
        "../evil.txt",
        "..\\evil.txt",
        "/../evil.txt",
        "C:\\evil.txt",
        "\\\\server\\share\\evil.txt",
    ],
)
def test_kwaj_safe_extract_blocks_traversal(tmp_path, name):
    kwj = KwajFile(str(FIXTURES / "sample.kwj"))
    with pytest.raises(KwajPathTraversalError):
        kwj.extract(str(tmp_path), out_name=name, safe=True)



from __future__ import annotations

import os
import posixpath
import re


_DRIVE_RE = re.compile(r"^[A-Za-z]:")


def normalize_member_path(name: str) -> str:
    """Normalize an archive member path for safe extraction."""
    if "\x00" in name:
        raise ValueError("NUL byte in member name")
    raw = name.replace("\\", "/")
    if _DRIVE_RE.match(raw):
        raise ValueError("Drive-letter paths are not allowed")
    if raw.startswith("//"):
        raise ValueError("UNC paths are not allowed")
    if raw.startswith("/"):
        raise ValueError("Absolute paths are not allowed")
    norm = posixpath.normpath(raw)
    if norm in {".", ""}:
        raise ValueError("Empty member path")
    if norm.startswith("../") or norm == "..":
        raise ValueError("Path traversal is not allowed")
    if norm.startswith("/"):
        raise ValueError("Absolute paths are not allowed")
    return norm


def safe_join(dest_dir: str, name: str) -> str:
    norm = normalize_member_path(name)
    dest_dir_abs = os.path.abspath(dest_dir)
    target = os.path.abspath(os.path.join(dest_dir_abs, *norm.split("/")))
    if os.path.commonpath([dest_dir_abs, target]) != dest_dir_abs:
        raise ValueError("Path traversal is not allowed")
    return target


def unsafe_join(dest_dir: str, name: str) -> str:
    parts = name.replace("\\", "/").split("/")
    return os.path.join(dest_dir, *parts)


def ensure_parent(path: str) -> None:
    parent = os.path.dirname(path)
    if parent:
        os.makedirs(parent, exist_ok=True)



from pathlib import Path

import pytest

from pylibmspack import (
    ChmArchive,
    ChmError,
    ChmFormatError,
    ChmPathTraversalError,
)


FIXTURES = Path(__file__).resolve().parent / "fixtures"
SAMPLE_CHM = FIXTURES / "sample.chm"

if not SAMPLE_CHM.exists():
    pytest.skip("sample.chm fixture not available", allow_module_level=True)


def test_chm_list_and_info():
    chm = ChmArchive(str(SAMPLE_CHM))
    info = chm.info()
    for key in [
        "filename",
        "length",
        "version",
        "timestamp",
        "language",
        "dir_offset",
        "num_chunks",
        "chunk_size",
        "density",
        "depth",
        "index_root",
        "first_pmgl",
        "last_pmgl",
        "files_count",
        "sysfiles_count",
    ]:
        assert key in info
    assert info["files_count"] >= 1
    assert info["length"] > 0
    assert info["chunk_size"] > 0
    assert info["num_chunks"] >= 0

    files = chm.files(include_system=False)
    assert files
    entry = files[0]
    assert entry["size"] >= 0
    assert "section" in entry
    assert "section_id" in entry


def test_chm_read_and_extract(tmp_path):
    chm = ChmArchive(str(SAMPLE_CHM))
    files = chm.files(include_system=False)
    files = [f for f in files if f["size"] > 0]
    smallest = min(files, key=lambda f: f["size"])
    data = chm.read(smallest["name"], max_size=smallest["size"])
    assert len(data) == smallest["size"]

    out_path = chm.extract(smallest["name"], str(tmp_path))
    assert Path(out_path).read_bytes() == data


def test_chm_from_bytes():
    data = SAMPLE_CHM.read_bytes()
    chm = ChmArchive.from_bytes(data)
    files = chm.files(include_system=False)
    files = [f for f in files if f["size"] > 0]
    smallest = min(files, key=lambda f: f["size"])
    payload = chm.read(smallest["name"], max_size=smallest["size"])
    assert len(payload) == smallest["size"]


def test_chm_invalid_bytes():
    chm = ChmArchive.from_bytes(b"not a chm")
    with pytest.raises(ChmFormatError):
        chm.info()


def test_chm_missing_member():
    chm = ChmArchive(str(SAMPLE_CHM))
    with pytest.raises(ChmError):
        chm.read("no_such_member.txt")


@pytest.mark.parametrize(
    "name",
    [
        "../evil.txt",
        "..\\evil.txt",
        "/../evil.txt",
        "C:\\evil.txt",
        "\\\\server\\share\\evil.txt",
    ],
)
def test_chm_safe_extract_blocks_traversal(tmp_path, name):
    chm = ChmArchive(str(SAMPLE_CHM))
    with pytest.raises(ChmPathTraversalError):
        chm.extract(name, str(tmp_path), safe=True)



from __future__ import annotations

import os
import tempfile
from typing import Optional, TypedDict

from . import _cab
from ._paths import ensure_parent, safe_join, unsafe_join
from .errors import (
    ChmDecompressionError,
    ChmError,
    ChmFormatError,
    ChmPathTraversalError,
)


_ERR_OK = getattr(_cab, "MSPACK_ERR_OK", 0)
_ERR_DATAFORMAT = getattr(_cab, "MSPACK_ERR_DATAFORMAT", -1)
_ERR_DECRUNCH = getattr(_cab, "MSPACK_ERR_DECRUNCH", -1)
_ERR_BADCOMP = getattr(_cab, "MSPACK_ERR_BADCOMP", -1)
_ERR_SIGNATURE = getattr(_cab, "MSPACK_ERR_SIGNATURE", -1)
_ERR_CHECKSUM = getattr(_cab, "MSPACK_ERR_CHECKSUM", -1)
_ERR_READ = getattr(_cab, "MSPACK_ERR_READ", -1)


def _raise_for_err(err: int, context: str) -> None:
    if err == _ERR_OK:
        return
    if err in {_ERR_DATAFORMAT, _ERR_BADCOMP, _ERR_SIGNATURE, _ERR_CHECKSUM, _ERR_READ}:
        raise ChmFormatError(f"{context} failed: libmspack error {err}")
    if err == _ERR_DECRUNCH:
        raise ChmDecompressionError(f"{context} failed: libmspack error {err}")
    raise ChmError(f"{context} failed: libmspack error {err}")


def _safe_join(dest_dir: str, name: str) -> str:
    try:
        return safe_join(dest_dir, name)
    except ValueError as exc:
        raise ChmPathTraversalError(str(exc)) from exc


def _unsafe_join(dest_dir: str, name: str) -> str:
    return unsafe_join(dest_dir, name)


def _ensure_parent(path: str) -> None:
    ensure_parent(path)


def _normalize_chm_name(name: str) -> str:
    # CHM members frequently start with a leading '/' that is not an absolute path.
    return name.lstrip("/")


class ChmFileInfo(TypedDict):
    """Typed metadata for a CHM member."""

    name: str
    size: int
    offset: int
    section_id: int
    section: str
    is_system: bool


class ChmInfo(TypedDict):
    """Typed metadata for a CHM header."""

    filename: Optional[str]
    length: int
    version: int
    timestamp: int
    language: int
    dir_offset: int
    num_chunks: int
    chunk_size: int
    density: int
    depth: int
    index_root: int
    first_pmgl: int
    last_pmgl: int
    files_count: int
    sysfiles_count: int


class ChmArchive:
    """Read and extract files from a CHM archive.

    Parameters
    ----------
    path:
        Path to a CHM file on disk.
    """

    def __init__(self, path: str) -> None:
        self.path = os.fspath(path)
        self._data: Optional[bytes] = None

    @classmethod
    def from_bytes(cls, data: bytes) -> "ChmArchive":
        """Create an archive backed by an in-memory CHM buffer."""
        obj = cls.__new__(cls)
        obj.path = "<memory>"
        obj._data = bytes(data)
        return obj

    def files(self, *, include_system: bool = True) -> list[ChmFileInfo]:
        """Return metadata for each member in the CHM.

        Parameters
        ----------
        include_system:
            Include system (\"::\") entries when True.
        """
        if self._data is not None:
            files, err = _cab.chm_list_files_bytes(self._data)
        else:
            files, err = _cab.chm_list_files(self.path)
        _raise_for_err(err, "listing")
        if files is None:
            return []
        if include_system:
            return files
        return [entry for entry in files if not entry.get("is_system")]

    def info(self) -> ChmInfo:
        """Return parsed CHM header information."""
        if self._data is not None:
            info, err = _cab.chm_info_bytes(self._data)
        else:
            info, err = _cab.chm_info(self.path)
        _raise_for_err(err, "info")
        if info is None:
            raise ChmError("info failed: no data")
        return info

    def read(self, name: str, *, max_size: int = 256 * 1024 * 1024) -> bytes:
        """Extract a member and return its bytes.

        Parameters
        ----------
        name:
            Member name within the CHM.
        max_size:
            Maximum allowed size in bytes. Defaults to 256 MiB.
        """
        if max_size <= 0:
            raise ValueError("max_size must be positive")
        member = _normalize_chm_name(name)
        with tempfile.TemporaryDirectory(prefix="pylibmspack-") as tmp:
            out_path = _safe_join(tmp, member)
            _ensure_parent(out_path)
            if self._data is not None:
                err = _cab.chm_extract_file_bytes(self._data, name, out_path)
            else:
                err = _cab.chm_extract_file(self.path, name, out_path)
            _raise_for_err(err, "extract")
            size = os.path.getsize(out_path)
            if size > max_size:
                raise ChmError(f"member exceeds max_size ({size} > {max_size})")
            with open(out_path, "rb") as f:
                return f.read()

    def extract(self, name: str, dest_dir: str, *, safe: bool = True) -> str:
        """Extract a member to disk and return the output path."""
        member = _normalize_chm_name(name)
        out_path = _safe_join(dest_dir, member) if safe else _unsafe_join(dest_dir, member)
        _ensure_parent(out_path)
        if self._data is not None:
            err = _cab.chm_extract_file_bytes(self._data, name, out_path)
        else:
            err = _cab.chm_extract_file(self.path, name, out_path)
        _raise_for_err(err, "extract")
        return out_path

    def extract_all(self, dest_dir: str, *, safe: bool = True, include_system: bool = True) -> list[str]:
        """Extract all members to disk and return output paths."""
        out_paths: list[str] = []
        for info in self.files(include_system=include_system):
            out_paths.append(self.extract(info["name"], dest_dir, safe=safe))
        return out_paths

    def extract_raw(self, name: str, dest_dir: str) -> str:
        """Extract a member using raw (unsafe) path handling."""
        return self.extract(name, dest_dir, safe=False)

    def extract_all_raw(self, dest_dir: str, *, include_system: bool = True) -> list[str]:
        """Extract all members using raw (unsafe) path handling."""
        return self.extract_all(dest_dir, safe=False, include_system=include_system)



from __future__ import annotations

import os
import tempfile
from typing import Optional, TypedDict

from . import _cab
from ._paths import ensure_parent, safe_join, unsafe_join
from .errors import (
    SzddDecompressionError,
    SzddError,
    SzddFormatError,
    SzddPathTraversalError,
)


_ERR_OK = getattr(_cab, "MSPACK_ERR_OK", 0)
_ERR_DATAFORMAT = getattr(_cab, "MSPACK_ERR_DATAFORMAT", -1)
_ERR_DECRUNCH = getattr(_cab, "MSPACK_ERR_DECRUNCH", -1)
_ERR_BADCOMP = getattr(_cab, "MSPACK_ERR_BADCOMP", -1)
_ERR_SIGNATURE = getattr(_cab, "MSPACK_ERR_SIGNATURE", -1)
_ERR_CHECKSUM = getattr(_cab, "MSPACK_ERR_CHECKSUM", -1)
_ERR_READ = getattr(_cab, "MSPACK_ERR_READ", -1)


def _raise_for_err(err: int, context: str) -> None:
    if err == _ERR_OK:
        return
    if err in {_ERR_DATAFORMAT, _ERR_BADCOMP, _ERR_SIGNATURE, _ERR_CHECKSUM, _ERR_READ}:
        raise SzddFormatError(f"{context} failed: libmspack error {err}")
    if err == _ERR_DECRUNCH:
        raise SzddDecompressionError(f"{context} failed: libmspack error {err}")
    raise SzddError(f"{context} failed: libmspack error {err}")


def _safe_join(dest_dir: str, name: str) -> str:
    try:
        return safe_join(dest_dir, name)
    except ValueError as exc:
        raise SzddPathTraversalError(str(exc)) from exc


def _unsafe_join(dest_dir: str, name: str) -> str:
    return unsafe_join(dest_dir, name)


def _ensure_parent(path: str) -> None:
    ensure_parent(path)


def _missing_char_str(value: int) -> str:
    if value == 0:
        return ""
    return bytes([value]).decode("latin-1", "strict")


def _apply_missing_char(name: str, missing: int) -> str:
    if not name:
        return name
    if missing == 0:
        return name
    ch = _missing_char_str(missing)
    if not ch:
        return name
    if name.endswith("_") or name.endswith("$"):
        return name[:-1] + ch
    return name[:-1] + ch


class SzddInfo(TypedDict):
    """Typed metadata for a SZDD header."""

    format_id: int
    format: str
    length: int
    missing_char: int
    missing_char_str: str
    suggested_name: str


class SzddFile:
    """Read and extract data from a SZDD-compressed file.

    Parameters
    ----------
    path:
        Path to a SZDD file on disk.
    """

    def __init__(self, path: str) -> None:
        self.path = os.fspath(path)
        self._data: Optional[bytes] = None
        self._name_override: Optional[str] = None

    @classmethod
    def from_bytes(cls, data: bytes, *, name: str = "memory.sz_") -> "SzddFile":
        """Create a SZDD reader backed by in-memory bytes."""
        obj = cls.__new__(cls)
        obj.path = "<memory>"
        obj._data = bytes(data)
        obj._name_override = name
        return obj

    def info(self) -> SzddInfo:
        """Return parsed SZDD header information."""
        if self._data is not None:
            info, err = _cab.szdd_info_bytes(self._data)
        else:
            info, err = _cab.szdd_info(self.path)
        _raise_for_err(err, "info")
        if info is None:
            raise SzddError("info failed: no data")
        missing = int(info.get("missing_char", 0))
        base_name = self._name_override or os.path.basename(self.path)
        suggested = _apply_missing_char(base_name, missing)
        info["missing_char_str"] = _missing_char_str(missing)
        info["suggested_name"] = suggested
        return info

    def suggested_name(self) -> str:
        """Return the default output filename based on the SZDD header."""
        return self.info()["suggested_name"]

    def read(self, *, max_size: int = 256 * 1024 * 1024) -> bytes:
        """Decompress and return file contents."""
        if max_size <= 0:
            raise ValueError("max_size must be positive")
        with tempfile.TemporaryDirectory(prefix="pylibmspack-") as tmp:
            out_name = self.suggested_name()
            if not out_name:
                out_name = "output"
            out_path = _safe_join(tmp, out_name)
            _ensure_parent(out_path)
            if self._data is not None:
                err = _cab.szdd_extract_bytes(self._data, out_path)
            else:
                err = _cab.szdd_extract(self.path, out_path)
            _raise_for_err(err, "extract")
            size = os.path.getsize(out_path)
            if size > max_size:
                raise SzddError(f"file exceeds max_size ({size} > {max_size})")
            with open(out_path, "rb") as f:
                return f.read()

    def extract(
        self,
        dest_dir: str,
        *,
        safe: bool = True,
        out_name: Optional[str] = None,
    ) -> str:
        """Decompress to disk and return the output path."""
        name = out_name or self.suggested_name()
        if not name:
            raise SzddError("output name is required")
        out_path = _safe_join(dest_dir, name) if safe else _unsafe_join(dest_dir, name)
        _ensure_parent(out_path)
        if self._data is not None:
            err = _cab.szdd_extract_bytes(self._data, out_path)
        else:
            err = _cab.szdd_extract(self.path, out_path)
        _raise_for_err(err, "extract")
        return out_path

    def extract_raw(self, dest_dir: str, *, out_name: Optional[str] = None) -> str:
        """Decompress using raw (unsafe) path handling."""
        return self.extract(dest_dir, safe=False, out_name=out_name)



from pathlib import Path

import pytest

from pylibmspack import SzddError, SzddFile, SzddFormatError


FIXTURES = Path(__file__).resolve().parent / "fixtures"
EXPECTED = FIXTURES / "expected"


def test_szdd_info_and_read(tmp_path):
    szdd = SzddFile(str(FIXTURES / "sample.tx_"))
    info = szdd.info()
    assert info["format"] in {"normal", "qbasic", "unknown"}
    assert info["length"] > 0
    assert info["suggested_name"] == "sample.txt"
    assert 0 <= info["missing_char"] <= 255

    data = szdd.read()
    assert data == (EXPECTED / "szdd.txt").read_bytes()

    out_path = szdd.extract(str(tmp_path))
    assert Path(out_path).read_bytes() == (EXPECTED / "szdd.txt").read_bytes()


def test_szdd_from_bytes():
    data = (FIXTURES / "sample.tx_").read_bytes()
    szdd = SzddFile.from_bytes(data, name="sample.tx_")
    info = szdd.info()
    assert info["suggested_name"] == "sample.txt"
    assert szdd.read() == (EXPECTED / "szdd.txt").read_bytes()


def test_szdd_invalid_bytes():
    szdd = SzddFile.from_bytes(b"not a szdd", name="bad.sz_")
    with pytest.raises(SzddFormatError):
        szdd.info()


def test_szdd_max_size():
    szdd = SzddFile(str(FIXTURES / "sample.tx_"))
    with pytest.raises(SzddError):
        szdd.read(max_size=1)



class MspackError(Exception):
    """Base error for libmspack handling."""


class MspackFormatError(MspackError):
    """File is corrupt, truncated, or uses unsupported features."""


class MspackDecompressionError(MspackError):
    """File could not be decompressed."""


class MspackPathTraversalError(MspackError):
    """Unsafe path detected during extraction."""


class CabError(MspackError):
    """Base error for CAB handling."""


class CabFormatError(CabError, MspackFormatError):
    """CAB file is corrupt, truncated, or uses unsupported features."""


class CabDecompressionError(CabError, MspackDecompressionError):
    """CAB file could not be decompressed."""


class CabPathTraversalError(CabError, MspackPathTraversalError):
    """Unsafe path detected during extraction."""


class ChmError(MspackError):
    """Base error for CHM handling."""


class ChmFormatError(ChmError, MspackFormatError):
    """CHM file is corrupt, truncated, or uses unsupported features."""


class ChmDecompressionError(ChmError, MspackDecompressionError):
    """CHM file could not be decompressed."""


class ChmPathTraversalError(ChmError, MspackPathTraversalError):
    """Unsafe path detected during extraction."""


class SzddError(MspackError):
    """Base error for SZDD handling."""


class SzddFormatError(SzddError, MspackFormatError):
    """SZDD file is corrupt, truncated, or uses unsupported features."""


class SzddDecompressionError(SzddError, MspackDecompressionError):
    """SZDD file could not be decompressed."""


class SzddPathTraversalError(SzddError, MspackPathTraversalError):
    """Unsafe path detected during extraction."""


class KwajError(MspackError):
    """Base error for KWAJ handling."""


class KwajFormatError(KwajError, MspackFormatError):
    """KWAJ file is corrupt, truncated, or uses unsupported features."""


class KwajDecompressionError(KwajError, MspackDecompressionError):
    """KWAJ file could not be decompressed."""


class KwajPathTraversalError(KwajError, MspackPathTraversalError):
    """Unsafe path detected during extraction."""
